/* dual-column.c - The "dual-column" edit format.
 *
 * Copyright (C) 2001  Oskar Liljeblad
 *
 * This file is part of the file renaming utilities (renameutils).
 *
 * This software is copyrighted work licensed under the terms of the
 * GNU General Public License. Please consult the file `COPYING' for
 * details.
 */

/* NOTE:
 * There will be major havoc if the indicators contain \n.
 */

#include <config.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "common/common.h"
#include "common/strbuf.h"
#include "common/string-utils.h"
#include "common/io-utils.h"
#include "common/intparse.h"
#include "common/tab-utils.h"
#include "common/error.h"
#include "common/memory.h"
#include "qmv.h"

static void reset_options();
static bool parse_options(char *options);
static char *option_generator(const char *text, int state);
static char *quote_first_last(char *old_name);
static void output(FILE *out, LList *files);
static inline void free_unread_lines();
static void unread_line(char *line);
static char *read_unread_line(FILE *in);
static bool input(FILE *in, LList *files);
static inline void swap_if_swap(char **a, char **b);
static bool process_dual_line(FileSpec *spec, char *line1, char *line2);
static bool process_single_line(FileSpec *spec, char *line);

static LList *unread_lines = NULL;	//FIXME
static int real_width;

/* Variables representing option */
static uint16_t width = 40;
static uint16_t tabsize = 8;
static bool swap = FALSE;
static bool use_spaces = FALSE;
static bool separate = FALSE;
static bool autowidth = TRUE;
static char *indicator1 = "";
static char *indicator2 = "";

EditFormat dual_column_format = {
	"dual-column",
	"dc",
	parse_options,
	option_generator,
	output,
	input
};

static char *option_array[] = {
	"help",
	"swap",
	"separate",
	"spaces",
	"tabsize",
	"width",
	"autowidth",
	"indicator1",
	"indicator2",
	NULL
};

enum {
	HELP_OPT,
	SWAP_OPT,
	SEPARATE_OPT,
	SPACES_OPT,
	TABSIZE_OPT,
	WIDTH_OPT,
	AUTOWIDTH_OPT,
	INDICATOR1_OPT,
	INDICATOR2_OPT,
};

#define NO_ARGUMENT() \
	if (value != NULL) { \
		warn("suboption `%s' for --option does not take an argument", option_array[c]); \
		reset_options(); \
		return FALSE; \
	}

#define REQUIRE_ARGUMENT() \
	if (value == NULL) { \
		warn("missing argument for --option suboption `%s'", option_array[c]); \
		reset_options(); \
		return FALSE; \
	}

static void
reset_options()
{
	swap = FALSE;
	separate = FALSE;
	use_spaces = FALSE;
	tabsize = 8;
	width = 40;
	autowidth = TRUE;
	indicator1 = "";
	indicator2 = "";
}

static bool
parse_options(char *options)
{
	reset_options();

	while (*options != '\0') {
		char *value;
		int c = getsubopt(&options, option_array, &value);

		if (c == -1) {
			warn("unknown suboption for --option: %s", value);
			reset_options();
			return FALSE;
		}

		switch (c) {
		case HELP_OPT:
			printf("\
Available options for `dual-column' format:\n\
\n\
  swap              swap location of old and new names when editing\n\
  separate          put a blank line between all renames\n\
  tabsize=SIZE      width of tabs in the edit file\n\
  spaces            indent with spaces only instead of tabs and spaces\n\
  width=WIDTH       column index which the second file name starts at\n\
  autowidth         enable automatic calculation of the width (default)\n\
  indicator1=TEXT   text to put before the first file name\n\
  indicator2=TEXT   text to put before the second file name\n\
\n\
Note that if a width is specified before autowidth, that width is used\n\
as a minimum width that the calculated width will never go below.\n");
			return FALSE;
		case SWAP_OPT:
			NO_ARGUMENT();
			swap = TRUE;
			break;
		case SEPARATE_OPT:
			NO_ARGUMENT();
			separate = TRUE;
			break;
		case SPACES_OPT:
			NO_ARGUMENT();
			use_spaces = TRUE;
			break;
		case TABSIZE_OPT:
			REQUIRE_ARGUMENT();
			if (!parse_uint16(value, &tabsize)) {
				warn("invalid number `%s' for tabsize suboption", value);
				reset_options();
				return FALSE;
			}
			break;
		case WIDTH_OPT:
			REQUIRE_ARGUMENT();
			if (!parse_uint16(value, &width)) {
				warn("invalid number `%s' for width suboption", value);
				reset_options();
				return FALSE;
			}
			autowidth = FALSE;
			break;
		case AUTOWIDTH_OPT:
			NO_ARGUMENT();
			autowidth = TRUE;
			break;
		case INDICATOR1_OPT:
			REQUIRE_ARGUMENT();
			indicator1 = value;
			break;
		case INDICATOR2_OPT:
			REQUIRE_ARGUMENT();
			indicator2 = value;
			break;
		}
	}

	return TRUE;
}

static char *
option_generator(const char *text, int state)
{
	static int c;
	if (state == 0)
		c = 0;
	while (option_array[c] != NULL) {
		char *name = option_array[c];
		c++;
		if (starts_with(name, text))
			return xstrdup(name);
	}
	return NULL;
}

static char *
quote_first_last(char *old_name)
{
	char *new_name;
	int c;
	int d;

	d = tab_index(old_name, real_width, tabsize);
	for (c = 0; c < d; c++) {
		if (old_name[c] != ' ' && old_name[c] != '\t')
			break;
	}

	new_name = strbuf_new();
	if (real_width != 0 && c == d)
		strbuf_append_char(&new_name, '\\');
	strbuf_append(&new_name, old_name);
	if (real_width != 0 && old_name[strlen(old_name)-1] == ' ')
		strbuf_insert_char(&new_name, -1, '\\');

	free(old_name);
	return strbuf_free_to_string(&new_name);
}

static void
output(FILE *out, LList *files)
{
	Iterator *it;
	LList *tmp = llist_new();
	int ilen = strlen(indicator1);
	char *old_name;
	char *new_name;

	real_width = width;

	/* Prepare list of old and new names */
	for (it = llist_iterator(files); iterator_has_next(it); ) {
		FileSpec *spec = iterator_next(it);

		old_name = quote_output_file(spec->old_name);
		old_name = quote_first_last(old_name);
		new_name = quote_output_file(spec->new_name);
		llist_add_last(tmp, old_name);
		llist_add_last(tmp, new_name);
		
		if (autowidth)
			real_width = MAX(real_width, strlen(swap ? new_name : old_name)+1);
	}
	iterator_free(it);

	while (!llist_is_empty(tmp)) {
		old_name = llist_remove_first(tmp);
		new_name = llist_remove_first(tmp);

		if (strlen(old_name) >= real_width+ilen) {
			fprintf(out, "%s%s\n", indicator1, old_name);
			tab_to(out, 0, real_width, (use_spaces ? 0 : tabsize));
			fprintf(out, "%s%s\n", indicator2, new_name);
		} else {
			fprintf(out, "%s%s", indicator1, old_name);
			tab_to(out, strlen(old_name)+ilen, real_width, (use_spaces ? 0 : tabsize));
			fprintf(out, "%s%s\n", indicator2, new_name);
		}
		if (separate && !llist_is_empty(tmp))
			fprintf(out, "\n");

		free(old_name);
		free(new_name);
	}

	llist_free(tmp);
}

static inline void
free_unread_lines()
{
	llist_iterate(unread_lines, (IteratorFunc) free);
	llist_free(unread_lines);
}

static void
unread_line(char *line)
{
	if (line != NULL)
		llist_add_first(unread_lines, line);
}

static char *
read_unread_line(FILE *in)
{
	if (!llist_is_empty(unread_lines))
		return llist_remove_first(unread_lines);
	return read_line(in);
}

static bool
input(FILE *in, LList *files)
{
	Iterator *it;
	char *line1;
	char *line2;

	unread_lines = llist_new();
	for (it = llist_iterator(files); iterator_has_next(it); ) {
		FileSpec *spec = iterator_next(it);
		int c, d;
		
		line1 = read_unread_line(in);
		if (line1 == NULL) {
			if (ferror(in))
				warn_errno("cannot read %s", edit_filename);
			else
				warn("premature end of %s looking for %s", edit_filename, spec->old_name);
			free_unread_lines();
			iterator_free(it);
			return FALSE;
		}

		line2 = read_unread_line(in);
		if (line2 == NULL) {
			if (ferror(in)) {
				warn_errno("cannot read %s", edit_filename);
				free(line1);
				free_unread_lines();
				iterator_free(it);
				return FALSE;
			}
			if (iterator_has_next(it)) {
				warn("premature end of %s looking for %s", edit_filename, spec->old_name);
				free(line1);
				free_unread_lines();
				iterator_free(it);
				return FALSE;
			}
			unread_line(line2);
			if (!process_single_line(spec, line1)) {
				free(line1);
				free_unread_lines();
				iterator_free(it);
				return FALSE;
			}
			free(line1);
		} else {
			d = tab_index(line2, real_width, tabsize);
			/* XXX: problem if indicator starts with at least width whitespace */
			for (c = 0; c < d; c++) {
				if (line2[c] != ' ' && line2[c] != '\t')
					break;
			}

			if (c != d) {
				unread_line(line2);
				if (!process_single_line(spec, line1)) {
					free(line1);
					free_unread_lines();
					iterator_free(it);
					return FALSE;
				}
				free(line1);
			} else {
				if (!process_dual_line(spec, line1, line2)) {
					free(line1);
					free(line2);
					free_unread_lines();
					iterator_free(it);
					return FALSE;
				}
				free(line1);
				free(line2);
			}
		}

		if (separate && iterator_has_next(it)) {
			char *line = read_unread_line(in);
			chomp(line);
			if (strlen(line) != 0) {
				warn("expected empty line, got %s", line);
				free(line);
				free_unread_lines();
				iterator_free(it);
				return FALSE;
			}
		}

		/*printf("< [%s]\n> [%s]\n", spec->old_name, spec->new_name);*/
	}
	iterator_free(it);

	line1 = read_unread_line(in);
	if (line1 != NULL) {
		chomp(line1);
		warn("expected end of file, got %s", line1);
		free(line1);
		free_unread_lines();
		return FALSE;
	}
	if (ferror(in)) {
		warn_errno("cannot read %s", edit_filename);
		free_unread_lines();
		return FALSE;
	}

	if (!llist_is_empty(unread_lines))
		internal_error("unread lines not empty; found %s\n", llist_get_first(unread_lines));

	free_unread_lines();

	return TRUE;
}

static inline void
swap_if_swap(char **a, char **b)
{
	if (swap) {
		char *t = *a;
		*a = *b;
		*b = t;
	}
}

static bool
process_dual_line(FileSpec *spec, char *line1, char *line2)
{
	int c;
	int d;
	char *old_name;
	char *new_name;

	if (!starts_with(line1, indicator1)) {
		warn("first name of %s did not start with an indicator", spec->old_name);
		return FALSE;
	}

	chomp(line1);
	chomp(line2);
	old_name = dequote_output_file(line1 + strlen(indicator1));
	string_strip_trailing(old_name, " \t");

	/* Check that characters up to the width' column are whitespace */
	d = tab_index(line2, real_width, tabsize);
	for (c = 0; c < d; c++) {
		if (line2[c] != '\t' && line2[c] != ' ') {
			warn("unrecognized character when space expected in %s", spec->old_name);
			return FALSE;
		}
	}
	if (!starts_with(line2+d, indicator2)) {
		warn("second name did not start with indicator in %s", spec->old_name);
		return FALSE;
	}
	new_name = dequote_output_file(line2 + d + strlen(indicator2));

	swap_if_swap(&old_name, &new_name);
	if (strcmp(spec->old_name, old_name) != 0) {
		warn("old name for %s does not match, is %s", spec->old_name, old_name);
		return FALSE;
	}
	free(old_name);

	spec->new_name = new_name;
	return TRUE;
}

static bool
process_single_line(FileSpec *spec, char *line)
{
	int c;
	int d;
	char *old_name;
	char *new_name;

	if (!starts_with(line, indicator1)) {
		warn("first name of %s did not start with an indicator", spec->old_name);
		return FALSE;
	}

	chomp(line);
	d = tab_index(line, real_width, tabsize);
	if (line[d] == '\0') {
		warn("destination name missing from line with %s", spec->old_name);
		return FALSE;
	}

	for (c = d-1; line[c] == ' ' || line[c] == '\t'; c--);
	if (c == d-1) {
		warn("expected whitespace at width-1");
		return FALSE;
	}
	if (line[c] == '\\')
		c++;

	line[c+1] = '\0';
	old_name = dequote_output_file(line + strlen(indicator1));

	if (!starts_with(line+d, indicator2)) {
		warn("second name of %s did not start with an indicator", spec->old_name);
		return FALSE;
	}
	new_name = dequote_output_file(line + d + strlen(indicator2));

	swap_if_swap(&old_name, &new_name);
	if (strcmp(spec->old_name, old_name) != 0) {
		warn("old name for %s does not match, is %s", spec->old_name, old_name);
		return FALSE;
	}
	free(old_name);

	spec->new_name = new_name;
	return TRUE;
}
